DECLARE FUNCTION IntersectSphere% (c1!, c0!)
DECLARE FUNCTION IntersectCone% (i%, n AS ANY)
DECLARE SUB IsConeHit (i%, Ns AS ANY)
DECLARE SUB loadpcx (a$, txnum%)
DECLARE SUB Precalculations ()
DECLARE SUB InitializeScene ()
DECLARE SUB IsSphereHit (spherenum%, Ns AS ANY, ds!, Left AS ANY, scanlinedir AS ANY)
DECLARE FUNCTION maxidx! (w AS ANY)
DECLARE SUB cross (V1 AS ANY, V2 AS ANY, v AS ANY)
DECLARE FUNCTION ceil% (x!)
DECLARE FUNCTION floor% (x!)
DECLARE SUB precalc ()
DECLARE SUB setpal2 ()
DECLARE FUNCTION totaldirectlight! (v0 AS ANY, v AS ANY, n AS ANY)
DECLARE FUNCTION phong! (L AS ANY, v AS ANY, n AS ANY)
DECLARE SUB setpal ()
DECLARE SUB raytrace (v0 AS ANY, dv AS ANY, c!, depth!)
DECLARE FUNCTION magnitude! (nix!, niy!, niz!)
DECLARE SUB normalize (x!, y!, z!)
DECLARE SUB rayscan ()
'==================================================================
'RAYTRACE H   Copyright 2000 Toshihiro Horie
'  changed to one-pass
'  * please do not redistribute this experimental source!
'  TODO: Fix inside of sphere bug
'  TODO: voxel space subdivision for speeding up recursive trace
'  TODO: implement cone intersection
'  TODO: real antialiasing on plane
'  -------------------------------------
'  8/29/2000
'  fixed plane overdraw bug,
'  added some fake antialiasing (blurring) to plane
'  note: this fake antialiasing intentially uses overdraw
'  fixed nonuniform texture dimness bug
'  fixed texture reflection banding bug
'
'  8/27/2000
'  fixed minor texture selection bug
'  fixed ATN bug in sphere texture mapping
'  added bilinear interpolation to above (looks wonderful!)
'  added cheap hack to simulate antialiasing on planes.
'  increased texture brightness by 60% or so
'  reduced reflectivity to 30%
'
'  8/22/2000
'  started cone intersection and scanline rejection calcs
'
'  8/11/2000
'  added texture mapping on spheres
'  added arbitrary planes
'  fixed elseif bug on intersection calcs
'  patched bug in totaldirect light where getting negative light
'  also added fast! pixel perfect scanline-interval sphere intersecter
'  based on raytrac5 code base, added refraction calcs (which don't work yet)
'  overflow bug patched in SUB phong.
'
'  6/22/2000 finally works
'  * Pasco is telling me that simply adding the light contributions
'  is not correct because it does not take into account the
'  color of the object.  There is no color yet-- I'm modelling
'  semi-perfect spherical mirrors with 20% light absoption and
'  a specular exponent of 82.
'  based on 5th attempt at a n-reflection ray tracer by Toshi Horie in Y2K.
'  The scene consists of:
'   * several spheres and point light sources
'==================================================================
CONST tracedepth = 3
CONST false = 0, true = NOT false
CONST dorefract = false
CONST dobilinear = true
CONST antialiasplane = false
CONST reflectivity = .4
CONST texturedimness = 101
CONST nrefract = 1.6               ' n2/n1 in physics literature
CONST invnrefract = 1! / nrefract  ' n1/n2 in physics literature
CONST scanwidth% = 320
CONST scanwidth1% = scanwidth% - 1
CONST NeverAgain = -2, OffScanLine = -1, Unknown = 0, Interval = 1, OnScanLine = 2

CONST pi = 3.14159265358979#
CONST ang180 = pi
CONST ang90 = pi / 2
CONST ang270 = 3 * pi / 2
CONST ang360 = pi * 2
CONST txmax = 127, tymax = 127
CONST tymaxover2 = tymax \ 2
CONST txscale = txmax / ang360

'actually, there is always one more object than specified here
CONST numlights = 1
CONST numspheres = 2
CONST numplanes = 0
CONST numcones = 1
CONST numtextures = 1


COMMON SHARED exitflag%
TYPE THeader            '= signs are for loading requirements
    MAN AS STRING * 1   'manufacturer = 10
    VER AS STRING * 1   'version = 5+
    ENC AS STRING * 1   'encoding = 1 (RLE)
    BPP AS STRING * 1   'bits per pixel per plane = 8
    MNX AS INTEGER      'minimum X size
    MNY AS INTEGER      'minimum Y size
    MAX AS INTEGER      'maximum X size
    MAY AS INTEGER      'maximum Y size
    HRS AS INTEGER      'horizontal resolution
    VRS AS INTEGER      'vertical resolution
    COL AS STRING * 48  'EGA color palette (ignored)
    RES AS STRING * 1   'reserved
    NMP AS STRING * 1   'number of planes = 1
    BPL AS INTEGER      'bytes per scanline
    PLI AS INTEGER      'palette info (ignored)
    FIL AS STRING * 58  'filler
END TYPE
TYPE rgbtype
    red AS INTEGER
    green AS INTEGER
    blue AS INTEGER
END TYPE
TYPE FHstatustype
    flag AS INTEGER
    IntervalL AS INTEGER
    IntervalR AS INTEGER
END TYPE
    
TYPE vec
    x AS SINGLE
    y AS SINGLE
    z AS SINGLE
END TYPE
TYPE spheretype
    a AS SINGLE 'center of sphere
    B AS SINGLE 'is (a,b,c)
    c AS SINGLE
    clr AS INTEGER 'base color of sphere
    R AS SINGLE  'radius of sphere
    r2 AS SINGLE 'radius^2
    da AS SINGLE
    db AS SINGLE
    dc AS SINGLE
    rn AS SINGLE
    textureofs AS INTEGER
    texturenum AS INTEGER
END TYPE
TYPE conetype
    v AS vec        'vertex (tip) of cone
    h AS vec        'axis of cone (from tip to base)
    hn AS vec        'axis of cone (from tip to base) normalized
    alpha AS SINGLE 'angle of spread of cone
    costheta AS SINGLE 'cosine of angle between scanline plane and cone axis
    cosalpha AS SINGLE 'cosine of alpha
    tantheta AS SINGLE
    tanalpha AS SINGLE
    tanalpha2 AS SINGLE
    hitflag AS INTEGER
END TYPE
TYPE pln
    a AS SINGLE 'coefficients for the plane equation
    B AS SINGLE ' Ax+By+Cz-D = 0
    c AS SINGLE
    D AS SINGLE
    S AS SINGLE 'used for speeding up plane intersection test
END TYPE
'center of projection (aka. the eye's position)
CLEAR , , 8000 'increase stack space cuz we're gonna do some major recursion

DIM SHARED eye AS vec
eye.x = 160: eye.y = 100: eye.z = -160
DIM SHARED o(numspheres) AS spheretype
DIM SHARED cone(numcones) AS conetype
DIM SHARED FHStatus(numspheres) AS FHstatustype
DIM SHARED plane(numplanes) AS pln
DIM SHARED lights(numlights) AS vec
DIM SHARED dx2%(-160 TO 160)
'$DYNAMIC
DIM SHARED texture%(0 TO txmax, 0 TO tymax, 0 TO numtextures)
'$STATIC

DIM SHARED NsLUT(0 TO 199) AS vec
DIM SHARED dsLUT(0 TO 199) AS SINGLE


CALL Precalculations
RANDOMIZE TIMER
CALL InitializeScene
PRINT "RAYTRACE G version 2.1 by Toshi"
PRINT "Loading texture maps..."
PRINT " " + CHR$(254) + " jupiter..."
loadpcx "jupiter2.pcx", 0
PRINT " " + CHR$(254) + " earth..."
loadpcx "earth2.pcx", 1

SCREEN 13


setpal2 'for grayscale
'setpal 'for false color

exitflag% = false
DO
    'create spheres of different sizes and textures
    FOR i% = 0 TO numspheres
        o(i%).a = RND * 640 - 100
        o(i%).B = RND * 400
        o(i%).c = i% * 100 + 80
        o(i%).clr = i% * 10 + 2
        o(i%).R = RND * 100 + 30
        o(i%).r2 = o(i%).R * o(i%).R
        o(i%).da = eye.x - o(i%).a
        o(i%).db = eye.y - o(i%).B
        o(i%).dc = eye.z - o(i%).c
        o(i%).rn = (o(i%).da * o(i%).da) + (o(i%).db * o(i%).db) + (o(i%).dc * o(i%).dc) - o(i%).R * o(i%).R
        o(i%).textureofs = INT(RND * txmax)
        o(i%).texturenum = INT(RND * 2.9999) - 1
    NEXT i%
    CLS
   
    starttime = TIMER
        CALL rayscan
    elapsed = TIMER - starttime
   
    t1 = TIMER: DO
        IF INKEY$ <> "" THEN END
    LOOP UNTIL TIMER - t1 > 2
LOOP UNTIL exitflag%
END

FUNCTION ceil% (x)
'IF x > 32767 THEN
'    x = 32767
'ELSEIF x < -32767 THEN
'    x = -32767
'END IF
ceil% = (x + .5) AND -1 '=INT(x + .999999!)
END FUNCTION

SUB cross (V1 AS vec, V2 AS vec, v AS vec)
'vector V = V1 x V2
v.x = V1.y * V2.z - V2.y * V1.z
v.y = V2.x * V1.z - V1.x * V2.z
v.z = V1.x * V2.y - V2.x * V1.y
END SUB

FUNCTION floor% (x)
'IF x > 32767 THEN
'    x = 32767
'ELSEIF x < -32767 THEN
'    x = -32767
'END IF
floor% = (x - .499999) AND -1
END FUNCTION

SUB InitializeScene

FOR i% = 0 TO numlights
    lights(i%).x = eye.x - 20 + (160 * i%)
    lights(i%).y = eye.y + 10 + (190 * i%) * -(i% AND 1)
    lights(i%).z = eye.z
NEXT i%

   
'plane eq: Ax+By+Cz=-D
FOR i% = 0 TO numplanes
SELECT CASE i%
CASE 0
    plane(i%).a = .8'.1
    plane(i%).B = 1
    plane(i%).c = 0
    plane(i%).D = eye.y + 50
    plane(i%).S = plane(i%).a * eye.x + plane(i%).B * eye.y + c * eye.z + plane(i%).D
CASE 2
    plane(i%).a = 0
    plane(i%).B = 0
    plane(i%).c = 1
    plane(i%).D = eye.z - 5000
    plane(i%).S = plane(i%).a * eye.x + plane(i%).B * eye.y + c * eye.z + plane(i%).D
CASE 1
    plane(i%).a = -.1
    plane(i%).B = 1
    plane(i%).c = 0
    plane(i%).D = eye.y + 50
    plane(i%).S = plane(i%).a * eye.x + plane(i%).B * eye.y + c * eye.z + plane(i%).D
CASE ELSE
    plane(i%).a = RND * 1
    plane(i%).B = RND * 1
    plane(i%).c = 0
    plane(i%).D = eye.y + RND * 100 + 60
    plane(i%).S = plane(i%).a * eye.x + plane(i%).B * eye.y + c * eye.z + plane(i%).D
END SELECT
NEXT i%

FOR i% = 0 TO numcones
    cone(i%).v.x = eye.x
    cone(i%).v.y = eye.y
    cone(i%).v.z = eye.z + 200
    cone(i%).h.x = 0
    cone(i%).h.y = -1
    cone(i%).h.z = 0
    cone(i%).hn.x = cone(i%).h.x
    cone(i%).hn.y = cone(i%).h.y
    cone(i%).hn.z = cone(i%).h.z
    CALL normalize(cone(i%).hn.x, cone(i%).hn.y, cone(i%).hn.z)
NEXT i%
END SUB

FUNCTION IntersectCone% (i%, n AS vec)
DIM u AS vec: DIM v AS vec: DIM w AS vec
'v=radius axis
v = cone(i%).hn
'u = v x n
CALL cross(v, n, u)
'w = u x v
CALL cross(u, v, w)

DIM d1 AS vec: DIM d2 AS vec
halfAB = SQR(cone(i%).tanalpha2 - cone(i%).tantheta * cone(i%).tantheta)
'common subexpression elimination using partial results
dpx = v.x + cone(i%).tantheta * w.x
dpy = v.y + cone(i%).tantheta * w.y
dpz = v.z + cone(i%).tantheta * w.z
del1.x = dpx + halfAB * u.x
del1.y = dpy + halfAB * u.y
del1.z = dpz + halfAB * u.z
del2.x = dpx - halfAB * u.x
del2.y = dpy - halfAB * u.y
del2.z = dpz - halfAB * u.z
'can use halfAB,tantheta, etc for cone normal (perpendicular vector) calc.
rscale = 1.2' FIXME
conen1u = halfAB * rscale
conen1v = -cone(i%).tanalpha2 * rscale
conen1w = cone(i%).tantheta * rscale
conen2u = -halfAB * rscale
conen2v = -cone(i%).tanalpha2 * rscale
conen2w = cone(i%).tantheta * rscale
DIM conen1 AS vec
DIM conen2 AS vec
conen1.x = conen1u * u.x + conen1v * v.x + conen1w * w.x
conen1.y = conen1u * u.y + conen1v * v.y + conen1w * w.y
conen1.z = conen1u * u.z + conen1v * v.z + conen1w * w.z
conen2.x = conen2u * u.x + conen2v * v.x + conen2w * w.x
conen2.y = conen2u * u.y + conen2v * v.y + conen2w * w.y
conen2.z = conen2u * u.z + conen2v * v.z + conen2w * w.z

'solve eye + t1*dv = cone(i%).V + r*del1
'solve eye + t2*dv = cone(i%).V + s*del2
'for t1 and t2
END FUNCTION

FUNCTION IntersectSphere% (c1, c0)
'intersectsphere% reports the number of distinct, real valued >1 intercepts
'INP: c1 and c0 are parameters for quadratic equation (A)*t^2+(C1)*t+C0 = 0
'OUT: c1 is used to return the parameter t values for the closest intercepts
'NOTE: USE EPSILON=1 BECAUSE THESE COORDINATES ARE NORMALIZED TO PIXEL UNITS!
eps = .1
discr = c1 * c1 - 4 * c0
IF discr < 0 THEN
    IntersectSphere% = 0
ELSEIF discr > .01 THEN
    sd = SQR(discr) 'always sd >=0
    c0 = .5 * (-c1 + sd)'order  !
    c1 = .5 * (-c1 - sd)'matters! because overwrites c1
   
    IF c0 <= eps AND c1 <= eps THEN IntersectSphere% = 0: EXIT FUNCTION
   
    'store closer intersection in c1, farther in c0
    IF c1 <= eps THEN c1 = c0
    IntersectSphere% = 2
ELSE 'discr = 0
    IntersectSphere% = 1
    c1 = -.5 * c1
END IF
END FUNCTION

SUB IsConeHit (i%, Ns AS vec)
'calculates at the beginning of the scanline
'whether cone i% is hit by any rays that pass through that scanline

'NOTE: NEED TO WRITE SPECIAL CASE WHEN RAY HITS the VERTEX of the CONE

ndothn = Ns.x * hn.x + Ns.y * hn.y + Ns.z * hn.z
cosphi = ndothn
IF cosphi > 1 OR cosphi < -1 THEN COLOR 255: PRINT "cosine too big": BEEP: END

cosalpha = COS(cone(i%).alpha)
costheta = SQR(1 - cosphi * cosphi)
cone(i%).cosalpha = cosalpha
cone(i%).costheta = costheta

IF costheta < cosalpha THEN
    'no intersections
    cone(i%).hitflag = OffScanLine
ELSEIF costheta > cosalpha THEN
    'up to two intersections
    cone(i%).hitflag = OnScanLine
    sintheta = -cosphi
    cone(i%).tantheta = sintheta / costheta
    tanalpha = TAN(cone(i%).alpha)
    cone(i%).tanalpha2 = tanalpha * tanalpha
ELSE 'grazing the cone so at most one intersection
    cone(i%).hitflag = OnScanLine
    sintheta = -cosphi'used cosine angle addition trig identity
    cone(i%).tantheta = sintheta / costheta
    tanalpha = TAN(cone(i%).alpha)
    cone(i%).tanalpha2 = tanalpha * tanalpha
END IF
END SUB

SUB IsSphereHit (spherenum%, Ns AS vec, ds, Left AS vec, scanlinedir AS vec) STATIC
'based on ideas and C code by Thomas Moller
'Purpose:
'   calculates whether sphere spherenum% will be hit on this scanline
'Params:
'   spherenum% - which sphere to use
'   Ns - normal of the scanline plane = (Right-Eye)x(Left-Eye)
'   ds - the offset of the scanline plane from the origin
'   Left - leftmost point on scanline
'   Scanlinedir - direction of scanline in space
'               = (dx,0,0) for our simple case
'   CONST scanwidth% - number of pixels per scanline
'Result:
'   fills FirstHitStatus() array
'throw away the coordinate
'that corresponds to max(Ns.x,Ns.y,Ns.z)
badidx% = maxidx(Ns)


signeddist = Ns.x * o(spherenum%).a + Ns.y * o(spherenum%).B + Ns.z * o(spherenum%).c + ds
IF signeddist > o(spherenum%).R THEN
    ' sphere is above the plane, so won't hit it anymore
    FHStatus(spherenum%).flag = NeverAgain
ELSEIF signeddist < -o(spherenum%).R THEN
    ' sphere is below the scanline plane
    FHStatus(spherenum%).flag = OffScanLine
ELSE
    'we might hit the sphere
    DIM D AS vec
    DIM h AS vec
    DIM Cc AS vec  'note: Cc is cut circle's center point
    DIM Nd AS vec
    '' vector CC = spherecenter - (signeddist * Ns )
    Cc.x = o(spherenum%).a - signeddist * Ns.x
    Cc.y = o(spherenum%).B - signeddist * Ns.y
    Cc.z = o(spherenum%).c - signeddist * Ns.z
    ''cr2=(Sr * Sr)- signeddist^2
    cr2 = o(spherenum%).R * o(spherenum%).R - signeddist * signeddist
    '' vector D = CC-Eyepos
    D.x = Cc.x - eye.x
    D.y = Cc.y - eye.y
    D.z = Cc.z - eye.z
    '' D2 = D dot D = (magnitude(D))^2
    d2 = D.x * D.x + D.y * D.y + D.z * D.z
    IF (d2 <= cr2) THEN
        FHStatus(spherenum%).flag = OnScanLine
        EXIT SUB
    END IF
    '' H = D cross Ns
    CALL cross(D, Ns, h)
    t1 = cr2 / d2
    sintheta = SQR(t1)
    costheta = SQR(1 - t1)
    SELECT CASE badidx%
        CASE 1
            t1 = h.y * sintheta
            t2 = D.y * costheta
            V1u = t1 + t2
            V2u = t2 - t1
            t1 = h.z * sintheta
            t2 = D.z * costheta
            V1v = t1 + t2
            V2v = t2 - t1
            t1 = Left.y - eye.y
            t2 = eye.z - Left.z
            t3 = V1u * scanlinedir.z - V1v * scanlinedir.y
            IF t3 <> 0! THEN
                'V1 is not parallel to scanline
                FHStatus(spherenum%).IntervalR = floor%((V1v * t1 + V1u * t2) / t3)
                t3 = V2u * scanlinedir.z - V2v * scanlinedir.y
                IF (t3 = 0!) THEN
                    FHStatus(spherenum%).IntervalL = 0
                ELSE
                    FHStatus(spherenum%).IntervalL = ceil%((V2v * t1 + V2u * t2) / t3)
                END IF
            ELSE 'rare case when V1 is parallel to scanline
                t3 = V2u * scanlinedir.z - V2v * scanlinedir.y
                FHStatus(spherenum%).IntervalL = ceil%((V2v * t1 + V2u * t2) / t3)
                FHStatus(spherenum%).IntervalR = scanwidth1%
            END IF
        CASE 2
        ' .x=.y from case 1
            t1 = h.x * sintheta
            t2 = D.x * costheta
            V1u = t1 + t2
            V2u = t2 - t1
            t1 = h.z * sintheta
            t2 = D.z * costheta
            V1v = t1 + t2
            V2v = t2 - t1
            t1 = Left.x - eye.x
            t2 = eye.z - Left.z
            t3 = V1u * scanlinedir.z - V1v * scanlinedir.x
            IF t3 <> 0! THEN
                'V1 is not parallel to scanline
                FHStatus(spherenum%).IntervalR = floor%((V1v * t1 + V1u * t2) / t3)
                t3 = V2u * scanlinedir.z - V2v * scanlinedir.x
                IF (t3 = 0!) THEN
                    FHStatus(spherenum%).IntervalL = 0
                ELSE
                    FHStatus(spherenum%).IntervalL = ceil%((V2v * t1 + V2u * t2) / t3)
                END IF
            ELSE 'rare case when V1 is parallel to scanline
                t3 = V2u * scanlinedir.z - V2v * scanlinedir.x
                FHStatus(spherenum%).IntervalL = ceil%((V2v * t1 + V2u * t2) / t3)
                FHStatus(spherenum%).IntervalR = scanwidth1%
            END IF
        CASE 3
            ' let .y=.z from case 2
            t1 = h.x * sintheta
            t2 = D.x * costheta
            V1u = t1 + t2
            V2u = t2 - t1
            t1 = h.y * sintheta
            t2 = D.y * costheta
            V1v = t1 + t2
            V2v = t2 - t1
            t1 = Left.x - eye.x
            t2 = eye.y - Left.y
            t3 = V1u * scanlinedir.y - V1v * scanlinedir.x
            IF t3 <> 0! THEN
                'V1 is not parallel to scanline
                FHStatus(spherenum%).IntervalR = floor%((V1v * t1 + V1u * t2) / t3)
                t3 = V2u * scanlinedir.y - V2v * scanlinedir.x
                IF (t3 = 0!) THEN
                    FHStatus(spherenum%).IntervalL = 0
                ELSE
                    FHStatus(spherenum%).IntervalL = ceil%((V2v * t1 + V2u * t2) / t3)
                END IF
            ELSE 'rare case when V1 is parallel to scanline
                t3 = V2u * scanlinedir.y - V2v * scanlinedir.x
                FHStatus(spherenum%).IntervalL = ceil%((V2v * t1 + V2u * t2) / t3)
                FHStatus(spherenum%).IntervalR = scanwidth1%
            END IF
        CASE ELSE
            SCREEN 0: PRINT "max(Ns) didn't work correctly": BEEP: END
    END SELECT
    ' check if interval is valid and set status
    IF ((FHStatus(spherenum%).IntervalL >= scanwidth%) OR (FHStatus(spherenum%).IntervalR < 0)) THEN
        FHStatus(spherenum%).flag = OffScanLine 'off screen to left or right
    ELSE
        IF (FHStatus(spherenum%).IntervalR >= scanwidth%) THEN FHStatus(spherenum%).IntervalR = scanwidth1%
        IF (FHStatus(spherenum%).IntervalL < 0) THEN FHStatus(spherenum%).IntervalL = 0
        IF ((FHStatus(spherenum%).IntervalL = 0) AND (FHStatus(spherenum%).IntervalR = scanwidth1%)) THEN
            FHStatus(spherenum%).flag = OnScanLine
        ELSE
            FHStatus(spherenum%).flag = Interval
        END IF
    END IF
    IF (FHStatus(spherenum%).IntervalL > FHStatus(spherenum%).IntervalR) THEN
        FHStatus(spherenum%).flag = OffScanLine
    END IF
END IF
'' this implicitly returns FHstatus
END SUB

DEFINT A-Z
SUB loadpcx (a$, idx)

DIM Byte AS STRING * 1    'used to read bytes
DIM Header AS THeader     'used for header info
DIM index AS LONG         'used to control decoding
DIM size AS LONG          'used to control decoding
DIM XSize AS INTEGER, YSize AS INTEGER 'image size
DIM totalbytes AS INTEGER 'used to measure scan lines
DIM value AS INTEGER      'used for converting char to int
DIM rlp                   'used for run length loop
DIM x AS LONG, y AS LONG  'used for pixel tracking

OPEN a$ FOR INPUT AS #1: CLOSE #1
OPEN a$ FOR BINARY AS #1
GET #1, 1, Header 'load header

'BASIC TEST (NOT fool-proof) -- REQUIREMENTS
IF ASC(Header.MAN) <> 10 THEN CLOSE #1: END 'not PCX
IF ASC(Header.VER) < 5 THEN CLOSE #1: END   'not 256 color
IF ASC(Header.ENC) <> 1 THEN CLOSE #1: END  'not RLE
IF ASC(Header.NMP) <> 1 THEN CLOSE #1: END  'not supportive of Mode X planes

'LOAD WHILE LOOP INFORMATION
XSize = Header.MAX - Header.MNX + 1 'load horizontal graphic size
YSize = Header.MAY - Header.MNY + 1 'load vertical graphic size
totalbytes = Header.BPL             'used to track end of scan line
size = 1& * XSize * YSize           'load size -- force long
IF size > 76800 THEN CLOSE : END    'graphic too large -- won't fit on screen



'CHANGE PALETTE
GET #1, LOF(1) - 768, Byte:         'load palette info
IF ASC(Byte) <> 12 THEN CLOSE : END 'No 256 color palette

FOR index = 0 TO 255
  GET #1, , Byte: red = ASC(Byte) \ 4'RED
  GET #1, , Byte: green = ASC(Byte) \ 4'GREEN
  GET #1, , Byte: blue = ASC(Byte) \ 4'BLUE
NEXT index
GET #1, 1, Header 'LOAD HEADER


'DECODE AND DISPLAY
WHILE index <= size                'While the picture isn't done loading...
    GET #1, , Byte                 'read a byte from the PCX file
    IF (ASC(Byte) AND &HC0) = &HC0 THEN 'test top two bits for 1'S (RLE?)
        rlp = ASC(Byte) AND &H3F           'set run length
        GET #1, , Byte                     'read data byte
        WHILE rlp > 0                'while run length has not yet finished
            IF NOT x > XSize THEN
                IF x <= txmax AND y <= tymax THEN
                texture%(x, y, idx) = ASC(Byte)'plot pixel
                END IF
            END IF
            x = x + 1                                'increment pixel position
            IF x >= totalbytes THEN x = 0: y = y + 1 'test for end of scanline
            rlp = rlp - 1                            'one less pixel to plot in run
            index = index + 1                        'increase total
        WEND
    ELSE             'if the byte is not encoded as RLE (then it's easy!)
        IF NOT x > XSize THEN
                IF x <= txmax AND y <= tymax THEN
                     texture%(x, y, idx) = ASC(Byte)'plot one pixel
                END IF
        END IF

        x = x + 1                                  'renew pixel position
        IF x >= totalbytes THEN x = 0: y = y + 1   'test for end of scanline
        index = index + 1                          'increase total
    END IF
WEND
CLOSE #1
END SUB

DEFSNG A-Z
FUNCTION magnitude (nix, niy, niz)
magnitude = SQR(nix * nix + niy * niy + niz * niz)
END FUNCTION

FUNCTION maxidx (w AS vec)
'Params:
'   W, the vector to find the max coordinate index of
'returns:
'   1 if W.x coordinate is greatest
'   2 if W.y coordinate is greatest
'   3 if W.z coordinate is greatest
                  MAX = ABS(w.x): tmaxidx = 1
IF (ABS(w.y) > MAX) THEN MAX = ABS(w.y): tmaxidx = 2
IF (ABS(w.z) > MAX) THEN MAX = ABS(w.z): tmaxidx = 3
maxidx = tmaxidx
END FUNCTION

SUB normalize (x, y, z)
mag = SQR(x * x + y * y + z * z)
'should make this beep in debug mode
IF mag = 0 THEN
    x = 0
    y = 0
    z = 0
    EXIT SUB
END IF
x = x / mag
y = y / mag
z = z / mag
END SUB

FUNCTION phong (L AS vec, v AS vec, n AS vec)
    DIM R AS vec
    fac = .8 / (numlights + 1) 'the object absorbs 20% of the light that shines on it.
    m = 82  'specularity coefficient (higher is more specular (and shinier) )
    'R is reflection of L about N
    'R = 2(N dot L)*N - L
    B = (n.x * L.x + n.y * L.y + n.z * L.z)
    ndotL2 = B * 2
    R.x = ndotL2 * n.x - L.x
    R.y = ndotL2 * n.y - L.y
    R.z = ndotL2 * n.z - L.z

    'brightness is proportional to (R dot V)^m
     rdotv = (R.x * v.x + R.y * v.y + R.z * v.z)
     IF rdotv <= 0 THEN
        'rdotv=0
        phong = (B * 125) * fac: EXIT FUNCTION
     ELSE
        'use m=8 for a somewhat shiny surface
        'use m=80 for very shiny surface
        'use m=256 for mirrors
        'Brightness = specular + diffuse + ambient
        IF rdotv > 1 THEN 'this was causing overflow
            phong = (127 + B * 125) * fac: EXIT FUNCTION
        ELSE
            phong = (rdotv ^ m * 127 + B * 125) * fac
        END IF
     END IF
END FUNCTION

SUB Precalculations

'nothing to precalc!

FOR i% = -160 TO 160
dx2%(i%) = i% * i%
NEXT i%

DIM BE AS vec 'vector Right-Eye
DIM AE AS vec 'vector Left-Eye
DIM Left AS vec  'leftmost pixel of scanline
DIM Right AS vec  'rightmost pixel of scanline
DIM Ns AS vec

Left.x = 0: Right.x = scanwidth1%
Left.z = z: Right.z = z
AE.z = Left.z - eye.z
BE.z = Right.z - eye.z
AE.x = Left.x - eye.x
BE.x = Right.x - eye.x

FOR y% = 0 TO 199
        Left.y = y%
        Right.y = y%
        AE.y = Left.y - eye.y
        BE.y = Right.y - eye.y
        CALL cross(BE, AE, Ns)
        CALL normalize(Ns.x, Ns.y, Ns.z)
        NsLUT(y%) = Ns
        dsLUT(y%) = -(Ns.x * eye.x + Ns.y * eye.y + Ns.z * eye.z)
NEXT y%
END SUB

SUB rayscan
'First pass of raytracing to determine visible surfaces.
'This calls the recursive raytrace routine.
CONST YFLIP = 199
DIM dv AS vec: DIM v0 AS vec
DIM n AS vec
DIM v AS vec

'for plane intersections
DIM pv AS vec
DIM pv0 AS vec
DIM pn AS vec
DIM Left AS vec

'added for fastSphereHit calcs
DIM Ns AS vec 'vector BExAE
Left.x = 0:
DIM scanlinedir AS vec
scanlinedir.x = 1
scanlinedir.y = 0
scanlinedir.z = 0

t1 = TIMER
'------------interlaced field 1-------------------------
DEF SEG = &HA000
FOR i% = 0 TO numspheres
    FHStatus(i%).flag = Unknown
NEXT i%
FOR i% = 0 TO numcones
    cone(i%).hitflag = Unknown
NEXT i%

Left.z = z 'don't erase

dz = z - eye.z
dz2 = dz * dz
clr = 0
'pp% = 320 * 179
    FOR y% = 20 TO 179
        key$ = INKEY$
        IF key$ <> "" THEN END
        dy = y% - eye.y
        dy2 = dy * dy
        dyZ2 = dy2 + dz2
        Left.y = y% 'don't erase!
        ds = dsLUT(y%)
        Ns = NsLUT(y%)
        FOR i% = 0 TO numspheres
            IF FHStatus(i%).flag <> NeverAgain THEN
                CALL IsSphereHit(i%, Ns, ds, Left, scanlinedir)
                'IF FHStatus(i%).flag = Interval THEN
                '    LINE (FHStatus(i%).IntervalL, y%)-(FHStatus(i%).IntervalR, y%), 253
                'ELSEIF FHStatus(i%).flag = OnScanline THEN
                '    LINE (0, y%)-(scanwidth1%, y%), 128
                'END IF
            END IF
        NEXT i%
        FOR i% = 0 TO numcones
            CALL IsConeHit(i%, Ns)
            ' don't use this as a hittest
            ' use only as a precalc at each scanline
        NEXT i%

        FOR x% = 0 TO scanwidth1%
            dx = x% - eye.x
            isdr2 = 1! / SQR(dx2%(dx) + dyZ2)
            dxn = dx * isdr2
            dyn = dy * isdr2
            dzn = dz * isdr2
          
            closestobj% = -1
            neart = 1E+09
            FOR i% = 0 TO numspheres
                IF (FHStatus(i%).flag = Interval AND x% >= FHStatus(i%).IntervalL AND x% <= FHStatus(i%).IntervalR) OR FHStatus(i%).flag = OnScanLine THEN
  
    'a is the quadratic, b is the linear, and c is the constant term
    'o(i%) is the center of the sphere we're testing against
               
                    'a = 1
                    B = 2 * (dxn * o(i%).da + dyn * o(i%).db + dzn * o(i%).dc)
                    c = o(i%).rn
                   
                    IF IntersectSphere(B, c) THEN
                        IF B < neart THEN
                            neart = B
                            closestobj% = i%
                        END IF
                    END IF
                END IF
            NEXT i%
           
            ' check for plane intersection
            IF closestobj% = -1 THEN
                closestpln% = -1
               
                neart2 = 1E+09
                'it hit none of the spheres
                'now test against  plane
                FOR i% = 0 TO numplanes
                    denom = plane(i%).a * dxn + plane(i%).B * dyn + plane(i%).c * dzn
                    IF denom <> 0 THEN
                        t = -(plane(i%).S) / denom
                        IF t > 0 THEN
                            IF t < neart2 THEN
                                closestpln% = i%
                                neart2 = t
                            END IF
                        END IF
                    END IF
                NEXT i%
            END IF

            IF closestpln% <> -1 AND neart2 < neart THEN
                pv.x = (neart2 * dxn)
                pv.y = (neart2 * dyn)
                pv.z = (neart2 * dzn)
                pv0.x = pv.x + eye.x
                pv0.y = pv.y + eye.y
                pv0.z = pv.z + eye.z
                pv.x = pv.x
                pv.y = pv.y
                pv.z = pv.z
                pn.x = plane(closestpln%).a
                pn.y = plane(closestpln%).B
                pn.z = plane(closestpln%).c
                'LOCATE 1, 1: COLOR 255: PRINT pv0.z
                IF pv0.z < 32766 THEN
                    zzz% = INT(pv0.z) AND 255
                ELSE
                    zzz% = RND * 3 + 128
                END IF
                zzz% = zzz% * totaldirectlight(pv0, pv, pn) \ 256
                'cheesy antialiasing
                IF antialiasplane THEN
                    PSET (x%, YFLIP - y%), (zzz% + POINT(x% - 1, YFLIP - y%)) \ 2
                    'POKE pp%, (zzz% + POINT(x% - 1, YFLIP - y%)) \ 2
                ELSE
                    PSET (x%, YFLIP - y%), (zzz%)
                    'POKE pp%, zzz%
                END IF
            END IF
           
            IF closestobj% <> -1 THEN
                'LOCATE 1, 1: COLOR 255: PRINT neart
                 'difference between eye and collision point
                 dxi = -neart * dxn
                 dyi = -neart * dyn
                 dzi = -neart * dzn

                 'position of collision in world coordinates
                 v0.x = eye.x - dxi
                 v0.y = eye.y - dyi
                 v0.z = eye.z - dzi
                 'V=(dxi,dyi,dzi) is the vector from sphere pt. back to eye
                 'new V0=(xi,yi,zi) is point of i'sect of ray on sphere's surface
                 'new ray origin is at collision point on surface of sphere
                 'COLOR 255: PRINT USING "delta=##.## ##.## ##.## ctr=##.## ##.## ##.##"; o(closestobj%).a; o(closestobj%).b; o(closestobj%).c

                'THESE TWO SHOULD BE EQUAL!
                ' COLOR 255: PRINT USING "###.#"; SQR((V0.x - o(closestobj%).a) ^ 2 + (V0.y - o(closestobj%).b) ^ 2 + (V0.z - o(closestobj%).c) ^ 2);
                ' PRINT USING "###.#"; o(closestobj%).r2
               
                'N = normal of the sphere at the collision point
                'N is normalized by the 1/r operation.
                 n.x = (v0.x - o(closestobj%).a) / o(closestobj%).R
                 n.y = (v0.y - o(closestobj%).B) / o(closestobj%).R
                 n.z = (v0.z - o(closestobj%).c) / o(closestobj%).R
               
                 'New direction dv = reflection of V about N.
                 v.x = dxi: v.y = dyi: v.z = dzi: CALL normalize(v.x, v.y, v.z)
                 ndoti2 = 2 * (n.x * v.x + n.y * v.y + n.z * v.z)
                
                 IF dorefract THEN
                    ndoti = .5 * ndoti2
                    IF ndoti2 >= 0! THEN
                        refr = nrefract
                    ELSE
                        'entering object from vaccuum
                        refr = invnrefract
                    END IF
                    disc2 = refr * refr * (ndoti * ndoti - 1) + 1
                    IF (disc2 > 0!) THEN
                        disc = SQR(disc2)
                        alpha = -refr * ndoti
                        IF (ndoti < 0!) THEN
                            alpha = alpha - disc
                        ELSE
                            alpha = alpha + disc
                        END IF
                        'note: changed v.x to -v.x and so on
                        dv.x = -v.x * refr + n.x * alpha
                        dv.y = -v.y * refr + n.y * alpha
                        dv.z = -v.z * refr + n.z * alpha
                    ELSE
                        'total internal refraction
                        PSET (x, y), 128: GOTO nextray
                    END IF
                 ELSE
                     'New direction dv = reflection of V about N.
                     dv.x = ndoti2 * n.x - v.x
                     dv.y = ndoti2 * n.y - v.y
                     dv.z = ndoti2 * n.z - v.z
                 END IF
                
                 'the new dv is normalized because symmetric
                 'reflections preserve length.

                 'add direct light contributions
                 tclr = totaldirectlight(v0, v, n)
                 clr = 0
                 IF tclr < 255 THEN
                    'recursively raytrace
                    depth = 0 'initialize ray tracer
                    CALL raytrace(v0, dv, clr, depth)
                 END IF
                 'reflectivity now defined as a global constant
                 clr = clr * reflectivity + tclr
                 IF clr > 255 THEN clr = 255
                 txnum% = o(closestobj%).texturenum
                 IF (txnum% >= 0) THEN
                     IF n.x = 0 THEN
                        txcoord = 0
                     ELSE
                         atnval = ATN(n.z / n.x)
                            IF n.x >= 0 THEN
                                IF n.z < 0 THEN
                                    txcoord = atnval + ang360
                                ELSE
                                    txcoord = atnval
                                END IF
                            ELSEIF n.x < 0 THEN
                                txcoord = atnval + ang180
                            END IF
                          
                     END IF
                     IF dobilinear THEN
                         txcoord = (txcoord * txscale + o(closestobj%).textureofs)
                         tycoord = tymaxover2 - n.y * (tymaxover2)
                         xrdn% = INT(txcoord)
                         yrdn% = INT(tycoord)
                         alphax = txcoord - xrdn%
                         alphay = tycoord - yrdn%
                         lite00 = texture%(xrdn% AND txmax, yrdn%, txnum%)
                         lite10 = texture%((xrdn% + 1) AND txmax, yrdn%, txnum%)
                         lite01 = texture%(xrdn% AND txmax, (yrdn% + 1) AND tymax, txnum%)
                         lite11 = texture%((xrdn% + 1) AND txmax, (yrdn% + 1) AND tymax, txnum%)
                         clrleft = (lite00 * (1 - alphay) + lite01 * (alphay))
                         clrright = (lite10 * (1 - alphay) + lite11 * (alphay))
                         clr = clr * (clrleft * (1 - alphax) + clrright * alphax) / texturedimness
                     ELSE
                        clr = clr * texture%(txcoord AND txmax, (tymaxover2 - n.y * (tymaxover2)), txnum%) / texturedimness
                     END IF
                     IF clr > 255 THEN clr = 255
                END IF
                 'Yes, you can use POKE here, but be careful!
                    PSET (x%, YFLIP - y%), clr
                    'POKE pp%, clr
            END IF
nextray:
            'pp% = pp% + 1
        NEXT
        'pp% = pp% - 640
    NEXT



'need to reset flags for every y pass
FOR mm% = 0 TO numspheres
    FHStatus(mm%).flag = Unknown
NEXT mm%

LOCATE 1, 1: COLOR 255: PRINT USING "##.##s  "; TIMER - t1
END SUB

SUB raytrace (v0 AS vec, dv AS vec, c, depth)
IF c >= 255 THEN EXIT SUB
'this is not quite working correctly
DIM v AS vec
DIM n AS vec

'for plane intersections
DIM pv AS vec
DIM pv0 AS vec
DIM pn AS vec

'PRE: v0 is new "eye" point for ray tracing
'     dv is normalized direction vector (direction eye is looking)
'base case
IF depth > tracedepth THEN EXIT SUB

'normal case
closestobj% = -1
neart = 9999
'check for intersection with all spheres
FOR i% = 0 TO numspheres
    diffx = v0.x - o(i%).a
    diffy = v0.y - o(i%).B
    diffz = v0.z - o(i%).c
    c1 = 2 * (dv.x * diffx + dv.y * diffy + dv.z * diffz)
    c0 = diffx * diffx + diffy * diffy + diffz * diffz - o(i%).r2
    
    IF IntersectSphere(c1, c0) THEN
        'this c1 is actually the t value at the intersection
        'between the sphere and our ray
        IF c1 < neart THEN
            neart = c1
            closestobj% = i%
        END IF
    END IF
NEXT i%

'
' should also check intersection with
' all planes and cones for general case
'

IF closestobj% <> -1 THEN
    'LOCATE 1, 1: COLOR 255: PRINT neart
     'difference between eye and collision point
     dxi = -neart * dv.x
     dyi = -neart * dv.y
     dzi = -neart * dv.z

     'position of collision in world coordinates
     v0.x = v0.x - dxi
     v0.y = v0.y - dyi
     v0.z = v0.z - dzi
     'V=(dxi,dyi,dzi) is the vector from sphere pt. back to eye
     'new V0=(xi,yi,zi) is point of i'sect of ray on sphere's surface
     'new ray origin is at collision point on surface of sphere
'     COLOR 255: PRINT USING "###.#"; SQR((V0.x - o(closestobj%).a) ^ 2 + (V0.y - o(closestobj%).b) ^ 2 + (V0.z - o(closestobj%).c) ^ 2);
'     PRINT USING "###.#"; o(closestobj%).r
   
    'N = normal of the sphere at the collision point
     n.x = (v0.x - o(closestobj%).a) / o(closestobj%).R
     n.y = (v0.y - o(closestobj%).B) / o(closestobj%).R
     n.z = (v0.z - o(closestobj%).c) / o(closestobj%).R
               
     'New direction dv = reflection of V about N.
     v.x = dxi: v.y = dyi: v.z = dzi: CALL normalize(v.x, v.y, v.z)
     ndoti2 = 2 * (n.x * v.x + n.y * v.y + n.z * v.z)

                 IF dorefract THEN
                    ndoti = .5 * ndoti2
                    IF ndoti2 >= 0! THEN
                        refr = nrefract
                    ELSE
                        'entering object from vaccuum
                        refr = invnrefract
                    END IF
                    disc2 = refr * refr * (ndoti * ndoti - 1) + 1
                    IF (disc2 > 0!) THEN
                        disc = SQR(disc2)
                        alpha = -refr * ndoti
                        IF (ndoti < 0!) THEN
                            alpha = alpha - disc
                        ELSE
                            alpha = alpha + disc
                        END IF
                        'note: changed v.x to -v.x and so on
                        dv.x = -v.x * refr + n.x * alpha
                        dv.y = -v.y * refr + n.y * alpha
                        dv.z = -v.z * refr + n.z * alpha
                    ELSE
                        'total internal refraction
                        PSET (x, y), 128: GOTO nextrayr1
                    END IF
                 ELSE
                     'New direction dv = reflection of V about N.
                     dv.x = ndoti2 * n.x - v.x
                     dv.y = ndoti2 * n.y - v.y
                     dv.z = ndoti2 * n.z - v.z
                 END IF

     tclr = totaldirectlight(v0, v, n)
     c = c + tclr * reflectivity


                 txnum% = o(closestobj%).texturenum
                 IF (txnum% >= 0) THEN
                     IF n.x = 0 THEN
                        txcoord = 0
                     ELSE
                         atnval = ATN(n.z / n.x)
                            IF n.x >= 0 THEN
                                IF n.z < 0 THEN
                                    txcoord = atnval + ang360
                                ELSE
                                    txcoord = atnval
                                END IF
                            ELSEIF n.x < 0 THEN
                                txcoord = atnval + ang180
                            END IF
                         
                     END IF

                     IF dobilinear THEN
                         txcoord = (txcoord * txscale + o(closestobj%).textureofs)
                         tycoord = tymaxover2 - n.y * (tymaxover2)
                         xrdn% = INT(txcoord)
                         yrdn% = INT(tycoord)
                         alphax = txcoord - xrdn%
                         alphay = tycoord - yrdn%
                         lite00 = texture%(xrdn% AND txmax, yrdn%, txnum%)
                         lite10 = texture%((xrdn% + 1) AND txmax, yrdn%, txnum%)
                         lite01 = texture%(xrdn% AND txmax, (yrdn% + 1) AND tymax, txnum%)
                         lite11 = texture%((xrdn% + 1) AND txmax, (yrdn% + 1) AND tymax, txnum%)
                         clrleft = (lite00 * (1 - alphay) + lite01 * (alphay))
                         clrright = (lite10 * (1 - alphay) + lite11 * (alphay))
                         c = c * (clrleft * (1 - alphax) + clrright * alphax) / 80
                     ELSE
                        c = c * texture%(txcoord AND txmax, (tymaxover2 - n.y * (tymaxover2)), txnum%) / texturedimness
                     END IF
                END IF

     IF c > 255 THEN EXIT SUB
     depth = depth + 1
     CALL raytrace(v0, dv, c, depth)
END IF
nextrayr1:

END SUB

SUB setpal
' ported from the C code in the book Computer Graphics by Foley et al.
numcolors = 256
MAX# = 63
DIM HSVPal(numcolors) AS rgbtype
SCREEN 13
S# = 0
S# = .9: 'saturation (0 to 1) 0=greyscale, 1=marker colors
v# = .9: 'brightness (0 to 1) 0=black, 1=brightest
p# = v# * (1# - S#)
conV# = numcolors / 360#
'if H was undefined, then we would have the greyscale.
FOR h% = 0 TO 359:       'hue (0 to 360) aka rainbow spectra
    c% = INT(h% * conV#): 'index into palette array
    hn# = h% / 60        'Hn is in the interval [0,6)
    i% = INT(hn#)        'i=floor(Hn)
    f# = hn# - i%        'fractional part of Hn
    q# = v# * (1# - (S# * f#))
    t# = v# * (1# - (S# * (1 - f#)))
    'PRINT i%, c%, q#, t#
    SELECT CASE i%
        CASE 0:
            HSVPal(c%).red = INT(v# * MAX#)
            HSVPal(c%).green = INT(t# * MAX#)
            HSVPal(c%).blue = INT(p# * MAX#)
        CASE 1:
            HSVPal(c%).red = INT(q# * MAX#)
            HSVPal(c%).green = INT(v# * MAX#)
            HSVPal(c%).blue = INT(p# * MAX#)
        CASE 2:
            HSVPal(c%).red = p# * MAX#
            HSVPal(c%).green = v# * MAX#
            HSVPal(c%).blue = t# * MAX#
        CASE 3:
            HSVPal(c%).red = INT(p# * MAX#)
            HSVPal(c%).green = INT(q# * MAX#)
            HSVPal(c%).blue = INT(v# * MAX#)
        CASE 4:
            HSVPal(c%).red = INT(t# * MAX#)
            HSVPal(c%).green = INT(p# * MAX#)
            HSVPal(c%).blue = INT(v# * MAX#)
        CASE 5:
            HSVPal(c%).red = INT(v# * MAX#)
            HSVPal(c%).green = INT(p# * MAX#)
            HSVPal(c%).blue = INT(q# * MAX#)
    END SELECT
NEXT h%
OUT &H3C8, 0: OUT &H3C9, 0: OUT &H3C9, 0: OUT &H3C9, 0
FOR i% = 1 TO numcolors - 1
    OUT &H3C8, i%
    OUT &H3C9, HSVPal(i%).red
    OUT &H3C9, HSVPal(i%).green
    OUT &H3C9, HSVPal(i%).blue
NEXT i%
OUT &H3C8, 255: OUT &H3C9, 255: OUT &H3C9, 255: OUT &H3C9, 255
END SUB

SUB setpal2
'grayscale
OUT &H3C8, 0
FOR i% = 0 TO 255
    m% = i% * 3 \ 4
    red% = CINT(m% * .33)
    green% = CINT(m% * .36)
    IF green% > 63 THEN green% = 63
    blue% = CINT(m% * .31)
    OUT &H3C9, red%
    OUT &H3C9, green%
    OUT &H3C9, blue%
NEXT


END SUB

FUNCTION totaldirectlight (v0 AS vec, v AS vec, n AS vec)
'V is normalized view vector
'v0 is new eye point (point of reflection)
DIM R AS vec
DIM L AS vec
DIM dv AS vec
'for each light
clr = 0
FOR m% = 0 TO numlights
    blocked = false
    'L has the direction toward light from sphere
    L.x = lights(m%).x - v0.x
    L.y = lights(m%).y - v0.y
    L.z = lights(m%).z - v0.z
    'can get divide by zero or overflow if light is at reflection point.
    Lmag2 = L.x * L.x + L.y * L.y + L.z * L.z
    Linvmag2 = 1! / Lmag2
    lightdist = SQR(Lmag2)
    
    neart = 9.9E+09
    'check for intersection with all spheres
    FOR i% = 0 TO numspheres
        diffx = v0.x - o(i%).a
        diffy = v0.y - o(i%).B
        diffz = v0.z - o(i%).c
        c1 = 2 / lightdist * (L.x * diffx + L.y * diffy + L.z * diffz)
        c0 = diffx * diffx + diffy * diffy + diffz * diffz - o(i%).r2
        
        IF IntersectSphere(c1, c0) THEN
            'this c1 is actually the t value at the intersection
            'between the sphere and our ray
            IF c1 < neart THEN
                neart = c1
                'short circuit if light blocked by any sphere
                ' Can compare Lrealmag to neart because both neart
                '   and Lrealmag are in world coordinate units
                IF lightdist >= neart THEN blocked = true: EXIT FOR
            END IF
        END IF
    NEXT i%
    IF NOT blocked THEN
        'light was closest, so add its contribution
        'normalize L vector
        L.x = L.x / lightdist
        L.y = L.y / lightdist
        L.z = L.z / lightdist
        'now L is normalized
        clr = clr + phong(L, v, n)
    ELSE
        'actually, we should recursively trace the ray here,
        'but I assume than an eclipsed ray will not make a
        'significant contribution to the brightness
    END IF

NEXT m%
IF clr > 254 THEN clr = 254
'WEIRD THAT THIS IS NEEDED, MAYBE A BUG SOMEWHERE
'OR ROUNDOFF ERRORS...
IF clr < 0 THEN clr = 0

totaldirectlight = clr

END FUNCTION

